#include "clause.h"

clause_t *addNewClause(clauses_state_t *state, truth_value_t *assignment, int b1, int b2, int e, int *literals, int size)
{
    // Create new clause and add to lists add clause to state and check update list.
    if (state->clauseList.n + 1 == state->clauseList.maxN)
    {
        if (state->clauseList.maxN >= INT32_MAX / 2) // for rank<=5 should not be relevant
            EXIT_UNWANTED_STATE
        state->clauseList.maxN *= 2;
        state->clauseList.clauses = realloc(state->clauseList.clauses, sizeof(clause_t) * state->clauseList.maxN);
        CHECK_OUT_OF_MEMORY(state->clauseList.clauses)
        printf("c\tSize of clauses increased to %d\n", state->clauseList.maxN);
        fflush(stdout);
    }

    int index = state->clauseList.n;
    clause_t *clause = &(state->clauseList.clauses[index]);
    clause->b1 = b1; // TODO check if lit and not base indices.
    clause->b2 = b2;
    clause->e = e;
    clause->n = size;

    clause->indexFire = -1;
    clause->nFalse = 0;
    clause->nTrue = 0;

    for (int i = 0; i < size; i++)
    {
        clause->literals[i] = literals[i]; // copy list
        switch (assignment[literals[i]])
        {
        case truth_value_true:
            return NULL; // must not be added
        case truth_value_false:
            clause->nFalse++;
        default:
            break;
        }
    }

    clause->indexB1 = addElemToIntList(&state->baseLits2clauses[clause->b1], index);
    clause->indexB2 = addElemToIntList(&state->baseLits2clauses[clause->b2], index);

    for (int i = 0; i < size; i++) // add only the ones which are not assigned yet
    {
        if (assignment[literals[i]] == truth_value_free)
        {
            clause->indices[i] = addElemToIntList(&state->lit2clauses[literals[i]], index);
        }
        else // truth value false
        {
            clause->indices[i] = -1;
        }
    }

    if (clause->nFalse >= clause->n - 1 && clause->nTrue == 0)
    {
        clause->indexFire = addElemToIntList(&state->ready2fire, index);
    }

    state->clauseList.n++; // Only added at this point
    return clause;
}

// Update the clause state after adding the literal lit
void addLiteral(clauses_state_t *state, lit_t lit, truth_value_t *assignment)
{
    if (lit > 0)
    {
        int_list_t list = state->lit2clauses[ABS(lit)]; // get list of clauses which must be updated
        // first update all and then remove all from list, i.e., the list must be empty add the end.
        for (int i = 0; i < list.n; i++)
        {
            clause_t *cl = &state->clauseList.clauses[list.elems[i]];
            (cl->nTrue)++;
            if (cl->indexFire != -1)
            {
                int index2Remove = cl->indexFire;
                int_list_t *list = &(state->ready2fire);
                list->n--;
                if (index2Remove != list->n)
                {
                    list->elems[index2Remove] = list->elems[list->n];
                    clause_t *clause = &(state->clauseList.clauses[list->elems[index2Remove]]);
                    clause->indexFire = index2Remove;
                }

                cl->indexFire = -1;
            }
        }
    }

    if (lit < 0)
    {
        int_list_t list = state->lit2clauses[ABS(lit)]; // get list of clauses which must be updated
        // first update all and then remove all from list, i.e., the list must be empty add the end.
        for (int i = 0; i < list.n; i++)
        {
            clause_t *cl = &state->clauseList.clauses[list.elems[i]];
            (cl->nFalse)++;
            if (cl->nFalse >= cl->n - 1 && cl->nTrue == 0 && cl->indexFire == -1)
            {
                cl->indexFire = addElemToIntList(&state->ready2fire, list.elems[i]);
            }
        }
    }
}

// Update the clause state after removing the literal lit
void removeLiteral(clauses_state_t *state, lit_t lit, truth_value_t *assignment)
{
    if (lit > 0)
    {
        int_list_t list = state->lit2clauses[ABS(lit)]; // get list of clauses which must be updated
        // first update all and then remove all from list, i.e., the list must be empty add the end.
        for (int i = 0; i < list.n; i++)
        {
            clause_t *cl = &state->clauseList.clauses[list.elems[i]];
            (cl->nTrue)--;
            if (cl->nTrue == 0 && cl->nFalse == cl->n - 1) // add to ready2fire list
            {
                cl->indexFire = addElemToIntList(&state->ready2fire, list.elems[i]);
            }
        }
    }

    if (lit < 0)
    {
        int_list_t list = state->lit2clauses[ABS(lit)]; // get list of clauses which must be updated
        // first update all and then remove all from list, i.e., the list must be empty add the end.
        for (int i = 0; i < list.n; i++)
        {
            clause_t *cl = &state->clauseList.clauses[list.elems[i]];
            (cl->nFalse)--;
            if (cl->indexFire != -1 && cl->nFalse < cl->n - 1) // remove from ready to fire list
            {
                int index2Remove = cl->indexFire;
                int_list_t *list = &(state->ready2fire);
                list->n--;
                if (index2Remove != list->n)
                {
                    list->elems[index2Remove] = list->elems[list->n];
                    clause_t *clause = &(state->clauseList.clauses[list->elems[index2Remove]]);
                    clause->indexFire = index2Remove;
                }

                cl->indexFire = -1;
            }
        }
    }
}

/**
 * Remove clause from state and update all other values correspondingly, i.e., from update and ready2fire list.
 * 
 * For removing always similarlly pattern
 * 
 * Pattern for deleting from lists
 * 
 * int index2Remove = get the index from the clause
 * int_list_t *list = getTheList
 * list->n--;
 * if (index2Remove != list->n)
 * {
 *      list->elems[index2Remove] = list->elems[list->n]
 *      clause_t *cl = &(state->clauseList.clauses[list->elems[index2Remove]]);
 *      cl->fieldToUpdate = index2Remove;
 * }
 * 
 * 
 * Pattern for shifted clause with newIndex
 * 
 * int_list_t *list = getTheList
 * list[shiftedClause->ABC] = newIndex;
 */
void removeClause(clauses_state_t *state, int index)
{
    if (DEBUG)
        printf("Remove clause\n");
    clause_t clause = state->clauseList.clauses[index];

    // delete info from additional lists
    {
        int index2Remove = clause.indexB1;
        int_list_t *list = &(state->baseLits2clauses[clause.b1]);
        list->n--;
        if (index2Remove != list->n)
        {
            list->elems[index2Remove] = list->elems[list->n];
            clause_t *cl = &(state->clauseList.clauses[list->elems[index2Remove]]);
            if (cl->b1 == clause.b1) // some special care needed
                cl->indexB1 = index2Remove;
            else
                cl->indexB2 = index2Remove;
        }
    }
    {
        int index2Remove = clause.indexB2;
        int_list_t *list = &(state->baseLits2clauses[clause.b2]);
        list->n--;
        if (index2Remove != list->n)
        {
            list->elems[index2Remove] = list->elems[list->n];
            clause_t *cl = &(state->clauseList.clauses[list->elems[index2Remove]]);
            if (cl->b2 == clause.b2) // some special care needed
                cl->indexB2 = index2Remove;
            else
                cl->indexB1 = index2Remove;
        }
    }

    for (int i = 0; i < clause.n; i++)
    {
        if (clause.indices[i] != -1)
        {
            int index2Remove = clause.indices[i];
            int_list_t *list = &(state->lit2clauses[clause.literals[i]]);
            list->n--;
            if (index2Remove != list->n)
            {
                list->elems[index2Remove] = list->elems[list->n];
                clause_t *cl = &(state->clauseList.clauses[list->elems[index2Remove]]);

                for (int j = 0; j < cl->n; j++) // search for literal with same index
                {
                    if (cl->literals[j] == clause.literals[i])
                    {
                        cl->indices[j] = index2Remove;
                        break;
                    }
                }
            }
        }
    }

    // delete from ready to fire if present
    if (clause.indexFire != -1)
    {
        int index2Remove = clause.indexFire;
        int_list_t *list = &(state->ready2fire);
        list->n--;
        if (index2Remove != list->n)
        {
            list->elems[index2Remove] = list->elems[list->n];
            clause_t *cl = &(state->clauseList.clauses[list->elems[index2Remove]]);
            cl->indexFire = index2Remove;
        }
    }

    state->clauseList.n--;
    if (index != state->clauseList.n) // put last element from list to the index and update information
    {
        // Update the indices according to shift
        state->clauseList.clauses[index] = state->clauseList.clauses[state->clauseList.n];
        clause_t *shiftedClause = &(state->clauseList.clauses[index]);

        for (int i = 0; i < shiftedClause->n; i++)
        {
            if (shiftedClause->indices[i] != -1)
            {
                lit_t l = shiftedClause->literals[i];
                state->lit2clauses[l].elems[shiftedClause->indices[i]] = index;
            }
        }

        if (shiftedClause->indexFire != -1)
        {
            state->ready2fire.elems[shiftedClause->indexFire] = index;
        }

        state->baseLits2clauses[shiftedClause->b1].elems[shiftedClause->indexB1] = index;
        state->baseLits2clauses[shiftedClause->b2].elems[shiftedClause->indexB2] = index;
    }
}

// -----------------------------------------helper functions--------------------------------------

void resizeIntList(int_list_t *list)
{
    if (list == NULL)
    {
        EXIT_UNWANTED_STATE
    }
    if (list->maxN == 0)
    {
        list->elems = malloc(sizeof(int) * INITIAL_SIZE_INTEGER_LIST);
        CHECK_OUT_OF_MEMORY(list->elems);
        list->maxN = INITIAL_SIZE_INTEGER_LIST;
    }
    list->maxN *= 2;
    list->elems = realloc(list->elems, list->maxN * sizeof(int));
    CHECK_OUT_OF_MEMORY(list->elems);
}

// add elemnt on last position; allocate/resize list if necessary. Returns the index, where it is added.
int addElemToIntList(int_list_t *list, int elem)
{
    if (list->n == list->maxN)
        resizeIntList(list);
    list->elems[list->n++] = elem;
    return list->n - 1;
}

/**
 *  For debugging. Check if clause_state is conform with real assignment.
 *  Does not check if there are duplicated clauses (i.e., clauses with identical literals), or clauses missing.
 */
void checkState(clauses_state_t *state, truth_value_t *assignment, propagator_t *data)
{
    //if (DEBUG)
    printf("c\tCheck %d clauses\n", state->clauseList.n);
    // go over baseLits2clauses lists
    for (int i = 0; i < data->inst->solveLitMax; i++)
    {
        if (assignment[i] != truth_value_true) // not allowed to have clauses
        {
            if (state->baseLits2clauses[i].n != 0)
                EXIT_UNWANTED_STATE;
        }
        else
        {
            for (int j = 0; j < state->baseLits2clauses[i].n; j++)
            {
                if (state->baseLits2clauses[i].elems[j] >= state->clauseList.n)
                {
                    printf("erorlit: %d; %d >= %d\n", i, state->baseLits2clauses[i].elems[j], state->clauseList.n);
                    EXIT_UNWANTED_STATE
                }
                clause_t cl = state->clauseList.clauses[state->baseLits2clauses[i].elems[j]];

                if (cl.b1 != i && cl.b2 != i)
                {
                    printf("b1: %d, b2: %d, index: %d\n", cl.b1, cl.b2, i);

                    EXIT_UNWANTED_STATE
                }
            }
        }
    }

    // go over update lists
    for (int i = 0; i < data->inst->solveLitMax; i++)
    {
        for (int j = 0; j < state->lit2clauses[i].n; j++)
        {
            int index = state->lit2clauses[i].elems[j];
            if (index >= state->clauseList.n)
                EXIT_UNWANTED_STATE
            clause_t cl = state->clauseList.clauses[index];

            // atleast one of the literals must be the clause
            bool literalPresent = false;
            for (int k = 0; k < cl.n; k++)
            {
                if (cl.literals[k] == i)
                {
                    literalPresent = true;
                    if (cl.indices[k] != j)
                        EXIT_UNWANTED_STATE
                    break;
                }
            }
            if (!literalPresent)
            {
                printf("Lit %d, clause: ", i);
                for (int k = 0; k < cl.n; k++)
                {
                    printf("%d ", cl.literals[k]);
                }
                printf("\n");
                EXIT_UNWANTED_STATE
            }
        }
    }

    // go over ready to fire list
    for (int i = 0; i < state->ready2fire.n; i++)
    {
        if (state->ready2fire.elems[i] >= state->clauseList.n)
            EXIT_UNWANTED_STATE
        clause_t cl = state->clauseList.clauses[state->ready2fire.elems[i]];
        if (cl.indexFire != i)
            EXIT_UNWANTED_STATE

        if (cl.nFalse < cl.n - 1 || cl.nTrue != 0)
        {
            EXIT_UNWANTED_STATE
        }
    }

    // check each clause
    for (int i = 0; i < state->clauseList.n; i++)
    {
        clause_t cl = state->clauseList.clauses[i];

        if (state->baseLits2clauses[cl.b1].elems[cl.indexB1] != i)
            EXIT_UNWANTED_STATE

        if (state->baseLits2clauses[cl.b2].elems[cl.indexB2] != i)
            EXIT_UNWANTED_STATE

        for (int j = 0; j < cl.n; j++)
        {
            if (cl.indices[j] != -1)
            {
                if (state->lit2clauses[cl.literals[j]].elems[cl.indices[j]] != i)
                    EXIT_UNWANTED_STATE
            }
        }

        if (cl.indexFire != -1)
        {
            if (state->ready2fire.elems[cl.indexFire] != i)
            {
                printf("%d %d \n", state->ready2fire.elems[cl.indexFire], i);
                EXIT_UNWANTED_STATE
            }
        }

        int nFalse = 0;
        int nTrue = 0;
        for (int j = 0; j < cl.n; j++)
        {
            lit_t lit = cl.literals[j];
            if (assignment[lit] == truth_value_false)
                nFalse++;
            if (assignment[lit] == truth_value_true)
                nTrue++;
        }

        if (cl.nFalse != nFalse)
        {
            printf("%d vs %d\n", cl.nFalse, nFalse);
            EXIT_UNWANTED_STATE
        }

        if (cl.nTrue != nTrue)
            EXIT_UNWANTED_STATE

        if ((cl.nFalse >= cl.n - 1 && cl.nTrue == 0) && cl.indexFire == -1)
        {
            printf("%d %d %d\n", cl.nFalse, cl.nTrue, cl.n);
            EXIT_UNWANTED_STATE
        }

        if (!(cl.nFalse >= cl.n - 1 && cl.nTrue == 0) && cl.indexFire != -1)
        {
            printf("%d %d %d\n", cl.nFalse, cl.nTrue, cl.n);
            EXIT_UNWANTED_STATE
        }
    }

    if (DEBUG)
        printf("Check completed\n");
}