#include "solver.h"
#include "util.h"
#include "clause.h"

#define DEBUG2 0
#define UNIT 0            // insert symmetry breaking clauses when unit
#define CLAUSE_VERSION2 0 //
#define ADD_BINARY 0      // add basis exchange property with only two options

bool extractForRec(int permutations[RANK][RANK], int *permutationOfBases, truth_value_t *assignment, clingo_propagate_control_t *control, propagator_t *data);
bool furtherPermutations(int permutationOfBases[RANK], int realPerm, int depth, int permutations[RANK][RANK], truth_value_t *assignment, clingo_propagate_control_t *control, propagator_t *data);
bool furtherPermutations2(int permutationOfBases[RANK], int realPerm, int pos, int depth, int permutations[RANK][RANK], truth_value_t *assignment, clingo_propagate_control_t *control, propagator_t *data, int state);

static inline void setminus(int set1[RANK], int set2[RANK], int result[RANK], int *nElems);
static inline int binom(int n, int k);
static inline int equalBase(int b1[RANK], int b2[RANK]);

void pavingRec(int elements[RANK], int depth, clingo_propagate_init_t *init, propagator_t *data, int *clause, int min);

static inline void printClause(int *clause, int n)
{
    for (int i = 0; i < n; i++)
        fprintf(arguments.clauseFile, "%d ", clause[i]);
    fprintf(arguments.clauseFile, "0");
}

int baseListLexBreak[N_BASES_SYM][RANK]; // Ordering for symmetry breaking. // TODO ensure selected base elements are at the beginning.

// Maps base to natural number (See https://math.stackexchange.com/questions/349924/function-mapping-combinations-to-natural-numbers)
// will be used to get solverlits and co for bases.
static inline int base2combinatorialIndex(int base[RANK]);
int index2Base[N_BASES][RANK]; // inverse function.

int fixedBaseIndices[RANK];

// TODO make with arguments and allow multithreading.
int numberOfInitialClauses = 0;

static inline int equalBase(int b1[RANK], int b2[RANK])
{
    for (int i = 0; i < RANK; i++)
        if (b1[i] != b2[i])
            return 0;
    return 1;
}

// sort after single element of an sorted array was changed given by index.
static inline void sortAfterChange(int arr[], int index, int n)
{
    if (index != 0 && arr[index - 1] > arr[index])
    {
        while (index != 0 && arr[index - 1] > arr[index])
        {
            SWAP(arr, index - 1, index)
            index--;
        }
    }
    else
    {
        while (index != n - 1 && arr[index] > arr[index + 1])
        {
            SWAP(arr, index, index + 1)
            index++;
        }
    }
}

int countCalls = 0;
int countEndNodes = 0;

/**
 *  Returns false if not maximal.
 *
 *  Permutations: permutations[i] gives permutation of the elements of the i-th initial base.
 *  realPerm: True if not the identity
 *  depth: For all i, permutations[i][j] is fixed for j < depth
 *  permDepth: For all i < permDepth, permutations[i][depth] is fixed.
 */
bool checkMaximility(int permutations[RANK][RANK], int realPerm, int depth, int permDepth, truth_value_t *assignment, clingo_propagate_control_t *control, propagator_t *data, int state)
{
    if (depth == RANK)
    {
        if (DEBUG2)
        {
            countEndNodes++;
            countCalls++;
        }

        int permutationOfBases[RANK]; // permutation of the intial bases
        for (int i = 0; i < RANK; i++)
            permutationOfBases[i] = i;

        // if (!furtherPermutations(permutationOfBases, realPerm, 0, permutations, assignment, control, data)) // TODO room for improvement in this function
        //     return false;

        return furtherPermutations2(permutationOfBases, realPerm, RANK, 0, permutations, assignment, control, data, state); // TODO room for improvement in this function
    }

    if (permDepth == RANK)
    {
        if (realPerm)
        {
            int base[RANK];
            for (int i = 0; i < RANK; i++)
                base[i] = i * RANK + depth;

            int basePerm[RANK];
            for (int i = 0; i < RANK; i++)
                basePerm[i] = permutations[base[i] / RANK][base[i] % RANK]; // indepenet of permutation of the originally fixed bases

            bubbleSort(basePerm, RANK);

            if (!equalBase(base, basePerm))
            {
                int index = base2combinatorialIndex(base);
                int indexPerm = base2combinatorialIndex(basePerm);

                int litOriginal = data->inst->baseIndex2LitSolver[index];
                int litPerm = data->inst->baseIndex2LitSolver[indexPerm];

                truth_value_t original = assignment[litOriginal];
                truth_value_t perm = assignment[litPerm];

                if (state == 0)
                {

                    // cases (u,1), (0,1) and cases (0,u)
                    if ((original != truth_value_true && perm == truth_value_true) || (original == truth_value_false && perm != truth_value_false))
                    {
                        return extractForRec(permutations, NULL, assignment, control, data);
                    }

                    // (1,0)
                    if (original == truth_value_true && perm == truth_value_false)
                    {
                        if (DEBUG2)
                            if (!(original == truth_value_true && perm == truth_value_false))
                                countEndNodes++;
                        return true; // is greater maximal
                    }

                    // (1,u) and cases (u,0), (u,u) ... other cases handled before
                    if (original == truth_value_free || perm == truth_value_free)
                    {
                        if (UNIT)
                            state = 1;
                        else
                            return true;
                    }

                    // cases left: both equal and not both free.
                }
                else if (UNIT)
                {

                    // cases (0,1)
                    if (original == truth_value_false && perm == truth_value_true)
                    {
                        return extractForRec(permutations, NULL, assignment, control, data);
                    }

                    // original 0 or perm 1 are the only cases where we can continue
                    if (!(original == truth_value_false || perm == truth_value_true))
                        return true;
                }
            }
        }

        // go into the recursion
        return checkMaximility(permutations, realPerm, depth + 1, 0, assignment, control, data, state);
    }

    for (int i = depth; i < RANK; i++)
    {
        SWAP(permutations[permDepth], i, depth);

        if (!checkMaximility(permutations, realPerm || i != depth, depth, permDepth + 1, assignment, control, data, state))
            return false;

        SWAP(permutations[permDepth], i, depth);
    }
    return true;
}

static inline int permuteSingleElement(int elem, int permutations[RANK][RANK], int *permutationOfBases)
{
    if (permutationOfBases == NULL)
        return permutations[elem / RANK][elem % RANK];
    else
        return permutations[permutationOfBases[elem / RANK]][elem % RANK];
}

/**
 * Given a base and permutations (permutations of elements and permutation of the initial bases)
 * Save the permutation of the given base in permBase
 */
static inline void permuteBase(int base[RANK], int permBase[RANK], int permutations[RANK][RANK], int *permutationOfBases)
{
    if (permutationOfBases == NULL)
    {
        for (int i = 0; i < RANK; i++)
            permBase[i] = permutations[base[i] / RANK][base[i] % RANK];
    }
    else
    {
        for (int i = 0; i < RANK; i++)
            permBase[i] = permutations[permutationOfBases[base[i] / RANK]][base[i] % RANK];
    }

    bubbleSort(permBase, RANK);
}

/**
 *  Returns false if not maximal
 *
 *  Given permutations[RANK][RANK] it additionally applies permutations of bases.
 */
bool furtherPermutations(int permutationOfBases[RANK], int realPerm, int depth, int permutations[RANK][RANK], truth_value_t *assignment, clingo_propagate_control_t *control, propagator_t *data)
{
    if (depth == RANK)
    {
        return !realPerm || extractForRec(permutations, permutationOfBases, assignment, control, data);
    }

    for (int i = depth; i < RANK; i++)
    {
        SWAP(permutationOfBases, i, depth)
        if (!furtherPermutations(permutationOfBases, realPerm || i != depth, depth + 1, permutations, assignment, control, data))
            return false;
        SWAP(permutationOfBases, i, depth)
    }

    return true;
}

/**
 *  Returns false if not maximal.
 *  More performant version
 *
 *  Given permutations[RANK][RANK] it additionally applies permutations of bases.
 */
bool furtherPermutations2(int permutationOfBases[RANK], int realPerm, int pos, int depth, int permutations[RANK][RANK], truth_value_t *assignment, clingo_propagate_control_t *control, propagator_t *data, int state)
{
    if (depth == RANK)
    {
        return !realPerm || extractForRec(permutations, permutationOfBases, assignment, control, data);
    }

    for (; pos < N_BASES_SYM; pos++)
    {
        if (baseListLexBreak[pos][RANK - 1] / RANK >= depth) // Check if already assigned
        {
            for (int i = depth; i < RANK; i++)
            {
                SWAP(permutationOfBases, depth, i);
                if (!furtherPermutations2(permutationOfBases, realPerm || (i != depth), pos, depth + 1, permutations, assignment, control, data, state))
                {
                    return false;
                }
                SWAP(permutationOfBases, depth, i);
            }
            return true;
        }

        if (!realPerm)
            continue;

        int *base = baseListLexBreak[pos]; // TODO check if the order by baseListLexBreak contains the "rows" first (i.e., the transversal bases)
        int permBase[RANK];
        permuteBase(base, permBase, permutations, permutationOfBases);

        int index = base2combinatorialIndex(base);
        int indexPerm = base2combinatorialIndex(permBase);

        if (index == indexPerm)
            continue;

        int litOriginal = data->inst->baseIndex2LitSolver[index];
        int litPerm = data->inst->baseIndex2LitSolver[indexPerm];

        truth_value_t original = assignment[litOriginal];
        truth_value_t perm = assignment[litPerm];

        if (!state)
        {

            // cases (u,1), (0,1), (0,u)
            if ((original != truth_value_true && perm == truth_value_true) ||
                (original == truth_value_false && perm != truth_value_false))
            {
                if (!extractForRec(permutations, permutationOfBases, assignment, control, data))
                    return false;
            }

            // cases (0,0)
            if (original == perm && original != truth_value_free)
                continue;

            if (UNIT)
            {
                // cases (1,u), (u,u)
                if ((original == truth_value_true && perm == truth_value_free) || (original == perm && original == truth_value_free))
                {
                    state = 1;
                    continue;
                }

                // (u,0)
                if (original == truth_value_free && perm == truth_value_false)
                {
                    state = 1;
                    continue;
                }
            }
        }
        else if (UNIT) // added so it can be discarded from compiler if not used
        {
            // cases (0,1)
            if (original == truth_value_false && perm == truth_value_true)
            {
                return extractForRec(permutations, permutationOfBases, assignment, control, data);
            }

            // cases (0,)
            if (original == truth_value_false || perm == truth_value_true)
                continue;
        }

        return true;
    }
    return true;
}

/**
 *  To avoid errors extract clause from give permutations an the truth assignemnt.
 *  return false if not maximal (if ADD_SINLGE clause is activated, then return false if conflicting)
 */
bool extractForRec(int permutations[RANK][RANK], int *permutationOfBases, truth_value_t *assignment, clingo_propagate_control_t *control, propagator_t *data)
{
    int state = 0;

    int clause[N_BASES_SYM * 2 + 2];
    int n = 0;

    for (int i = 0; i < N_BASES_SYM; i++)
    {
        int *base = baseListLexBreak[i]; // TODO check if the order by baseListLexBreak contains the "rows" first (i.e., the transversal bases)
        int permBase[RANK];
        permuteBase(base, permBase, permutations, permutationOfBases);

        int index = base2combinatorialIndex(base);
        int indexPerm = base2combinatorialIndex(permBase);

        if (index == indexPerm)
            continue;

        int litOriginal = data->inst->baseIndex2LitSolver[index];
        int litPerm = data->inst->baseIndex2LitSolver[indexPerm];

        truth_value_t original = assignment[litOriginal];
        truth_value_t perm = assignment[litPerm];

        if (!state)
        {

            // cases (u,1), (0,1), (0,u)
            if ((original != truth_value_true && perm == truth_value_true) ||
                (original == truth_value_false && perm != truth_value_false))
            {
                clause[n] = litOriginal;
                clause[n + 1] = -litPerm;

                bool res;
                if (!clingo_propagate_control_add_clause(control, clause, n + 2, clingo_clause_type_learnt, &res))
                    EXIT_UNWANTED_STATE
                if (arguments.clauseFile)
                {
                    printClause(clause, n + 2);
                    fprintf(arguments.clauseFile, ";s;[%d", permuteSingleElement(0, permutations, permutationOfBases));

                    for (int i = 1; i < ELEMENTS; i++)
                        fprintf(arguments.clauseFile, ",%d", permuteSingleElement(i, permutations, permutationOfBases));
                    fprintf(arguments.clauseFile, "]\n");
                }
                if (arguments.addSingleClause)
                    return false;
                else
                    return res;
            }

            // cases (0,0)
            if (original == perm && original == truth_value_false)
            {
                clause[n++] = litOriginal;
                continue;
            }

            // cases (1,1)
            if (original == perm && original == truth_value_true)
            {
                clause[n++] = -litPerm;
                continue;
            }

            if (UNIT)
            {
                // cases (1,u), (u,u)
                if ((original == truth_value_true && perm == truth_value_free) || (original == perm && original == truth_value_free))
                {
                    state = 1;
                    clause[n++] = -litPerm;
                    continue;
                }

                // (u,0)
                if (original == truth_value_free && perm == truth_value_false)
                {
                    state = 1;
                    clause[n++] = litOriginal;
                    continue;
                }
            }
        }
        else if (UNIT) // added so it can be discarded from compiler if not used
        {
            // cases (0,1)
            if (original == truth_value_false && perm == truth_value_true)
            {
                clause[n] = litOriginal;
                clause[n + 1] = -litPerm;

                bool res;
                if (!clingo_propagate_control_add_clause(control, clause, n + 2, clingo_clause_type_learnt, &res))
                    EXIT_UNWANTED_STATE
                if (arguments.clauseFile)
                {
                    printClause(clause, n + 2);
                    fprintf(arguments.clauseFile, ";s;[%d", permuteSingleElement(0, permutations, permutationOfBases));

                    for (int i = 1; i < ELEMENTS; i++)
                        fprintf(arguments.clauseFile, ",%d", permuteSingleElement(i, permutations, permutationOfBases));
                    fprintf(arguments.clauseFile, "]\n");
                }
                if (arguments.addSingleClause)
                    return false;
                else
                    return res;
            }

            // cases (0,)
            if (original == truth_value_false)
            {
                clause[n++] = litOriginal;
                continue;
            }

            // cases (,1)
            if (perm == truth_value_true)
            {
                clause[n++] = -litPerm;
                continue;
            }
        }
        return true;
    }
    return true;
}

// get to each (combinatorial) index the base, i.e., the inverse function.
void enumerateBases(int base[RANK], int depth, int minElement)
{
    if (depth == RANK)
    {
        int index = base2combinatorialIndex(base);
        for (int i = 0; i < RANK; i++)
            index2Base[index][i] = base[i];
        return;
    }

    for (int i = minElement; i < ELEMENTS; i++)
    {
        base[depth] = i;
        enumerateBases(base, depth + 1, i + 1);
    }
}

static inline int isElementInList(int *list, int elem, int n)
{
    for (int i = 0; i < n; i++)
        if (list[i] == elem)
            return 1;
    return 0;
}

static inline void printBasis(int *basis)
{
    fprintf(arguments.clauseFile, "[%d", basis[0]);
    for (int i = 1; i < RANK; i++)
    {
        fprintf(arguments.clauseFile, ",%d", basis[i]);
    }
    fprintf(arguments.clauseFile, "]");
}

/**
 * Precalculate some stuff, i.e., the fixed bases, the order of the lexicographic order, ...
 */
void init_static()
{
    int base[RANK];
    enumerateBases(base, 0, 0);

    int indicesAlreadyAdded[2 * RANK];

    int counter = 0;
    for (int i = 0; i < RANK; i++) // rows of the matrix
    {
        for (int j = 0; j < RANK; j++)
            baseListLexBreak[counter][j] = j * RANK + i;
        indicesAlreadyAdded[i] = base2combinatorialIndex(baseListLexBreak[counter]);
        counter++;
    }

    for (int i = 0; i < RANK; i++) // which columns of the matrix, i.e., fixed bases
    {
        int base[RANK];
        for (int j = 0; j < RANK; j++)
            base[j] = i * RANK + j;
        indicesAlreadyAdded[RANK + i] = base2combinatorialIndex(base);
    }

    // add some more bases except the rows and columns
    for (int i = 0; i < N_BASES; i++)
    {
        if (isElementInList(indicesAlreadyAdded, i, RANK * 2))
            continue;

        for (int j = 0; j < RANK; j++)
            baseListLexBreak[counter][j] = index2Base[i][j];
        counter++;
    }

    if (counter != N_BASES_SYM)
        EXIT_UNWANTED_STATE

    // init the fixed bases
    for (int i = 0; i < RANK; i++)
    {
        int base[RANK];
        for (int j = 0; j < RANK; j++)
            base[j] = RANK * i + j;

        fixedBaseIndices[i] = base2combinatorialIndex(base);
    }

    // for (int i = 0; i < N_BASES; i++)
    // {
    //     printf("%d:", i);
    //     for (int j = 0; j < RANK; j++)
    //         printf("%d ", index2Base[i][j]);
    //     if (base2combinatorialIndex(index2Base[i]) != i)
    //         EXIT_UNWANTED_STATE
    //     printf("\n");
    // }

    if (arguments.clauseFile)
    {
        fprintf(arguments.clauseFile, "[");
        printBasis(baseListLexBreak[0]);
        for (int i = 1; i < N_BASES_SYM; i++)
        {
            fprintf(arguments.clauseFile, ",");
            printBasis(baseListLexBreak[i]);
        }
        fprintf(arguments.clauseFile, "]\n");
    }
}

/**
 * Turns a clingo symbol into the corresponding base
 */
void symbol2Base(clingo_symbol_t sym, int base[RANK])
{
    clingo_symbol_t const *args;
    size_t nArgs;
    if (!clingo_symbol_arguments(sym, &args, &nArgs))
        EXIT_UNWANTED_STATE
    if (nArgs != RANK)
        EXIT_UNWANTED_STATE
    for (int i = 0; i < RANK; i++)
        if (!clingo_symbol_number(args[i], &base[i]))
            EXIT_UNWANTED_STATE
}

// Just prints the edges (Not very nice, but still gives the results)
void print_model(clingo_model_t const *model)
{
    clingo_symbol_t *atoms = NULL;
    size_t atoms_n;
    clingo_symbol_t const *it, *ie;
    char *str = NULL;
    size_t str_n = 0;

    // determine the number of (shown) symbols in the model
    if (!clingo_model_symbols_size(model, clingo_show_type_shown, &atoms_n))
        EXIT_UNWANTED_STATE
    // allocate required memory to hold all the symbols
    if (!(atoms = (clingo_symbol_t *)malloc(sizeof(*atoms) * atoms_n)))
        EXIT_UNWANTED_STATE
    // retrieve the symbols in the model
    if (!clingo_model_symbols(model, clingo_show_type_shown, atoms, atoms_n))
        EXIT_UNWANTED_STATE

    printf("Model:");
    for (it = atoms, ie = atoms + atoms_n; it != ie; ++it)
    {
        size_t n;
        char *str_new;
        // determine size of the string representation of the next symbol in the model
        if (!clingo_symbol_to_string_size(*it, &n))
            EXIT_UNWANTED_STATE
        if (str_n < n)
        {
            // allocate required memory to hold the symbol's string
            if (!(str_new = (char *)realloc(str, sizeof(*str) * n)))
                EXIT_UNWANTED_STATE
            str = str_new;
            str_n = n;
        }
        // retrieve the symbol's string
        if (!clingo_symbol_to_string(*it, str, n))
            EXIT_UNWANTED_STATE
        printf(" %s", str);
    }
    printf("\n");

    if (atoms)
        free(atoms);
    if (str)
        free(str);
    return;
}

/**
 * Starts solving for an inst
 */
void solve(rotta_inst_t *inst, inst_solving_t *solving)
{
    printf("c\tstart solving\n");

    solving->time = clock();
    inst->start = clock();

    int nFixed = solving->nPresentTuples + solving->nNotPresentTuples;
    int assumptions[nFixed];

    for (int i = 0; i < solving->nPresentTuples; i++)
    {
        int index = base2combinatorialIndex(solving->presentTuples[i]);
        int lit = inst->baseIndex2Lit[index]; // TODO check again if solver lits or normal lits. (But solver lits not known yet)
        assumptions[i] = lit;
    }

    for (int i = 0; i < solving->nNotPresentTuples; i++)
    {
        int index = base2combinatorialIndex(solving->notPresentTuples[i]);
        int lit = inst->baseIndex2Lit[index]; // TODO check again if solver lits
        assumptions[solving->nPresentTuples + i] = -lit;
    }

    // get a solve handle
    if (!clingo_control_solve(inst->control, clingo_solve_mode_yield | clingo_solve_mode_async, assumptions, nFixed, NULL, NULL, &(solving->handle)))
        EXIT_UNWANTED_STATE
    if (!clingo_solve_handle_resume(solving->handle))
        EXIT_UNWANTED_STATE

    clingo_solve_handle_wait(solving->handle, solving->timeout == 0 ? -1 : solving->timeout, &(solving->finished));
    if (!solving->finished)
    {
        if (!clingo_solve_handle_cancel(solving->handle))
            EXIT_UNWANTED_STATE
        goto solve_end;
    }

    if (!clingo_solve_handle_model(solving->handle, &(solving->model)))
        EXIT_UNWANTED_STATE

    // print the model
    if (solving->model)
    {
        printf("!!!!!!!!!!!!!!!SAT!!!!!!!!!!!!!!!\n");
        solving->sat = true;
        print_model(solving->model);
        EXIT_UNWANTED_STATE // Not expected that we get a model, but if we get one, we can return
    }
    else
        solving->sat = false;

solve_end:
    solving->time = clock() - solving->time;
    if (!clingo_solve_handle_close(solving->handle))
        EXIT_UNWANTED_STATE

    return;
}

// Called before each solving step, but only an effect at the first one
bool init_clingo(clingo_propagate_init_t *init, propagator_t *data)
{
    size_t threads = clingo_propagate_init_number_of_threads(init);
    if (data->states != NULL)
    {
        return true; // TODO check if several calls if number of threads has not changed
    }

    if (!(data->states = malloc(sizeof(state_t) * threads)))
        EXIT_UNWANTED_STATE
    memset(data->states, 0, sizeof(state_t) * threads);
    data->states_size = threads;

    // printf("c\tMapped literals: \n");
    // fflush(stdout);
    for (int i = 0; i < N_BASES; i++)
    {
        lit_t unmapped = data->inst->baseIndex2Lit[i];
        lit_t mapped;
        if (!clingo_propagate_init_solver_literal(init, unmapped, &mapped))
            EXIT_UNWANTED_STATE

        if (!clingo_propagate_init_add_watch(init, mapped))
            EXIT_UNWANTED_STATE
        if (!clingo_propagate_init_add_watch(init, -mapped))
            EXIT_UNWANTED_STATE

        // printf("c\tBefore mapping: %d After mapping %d\n", unmapped, mapped);

        if (mapped < 0)
            EXIT_UNWANTED_STATE

        data->inst->solverLit2baseIndex[mapped] = i;

        data->inst->baseIndex2LitSolver[i] = mapped;
        data->inst->solveLitMax = MAX(ABS(mapped), data->inst->solveLitMax);
        if (data->inst->solveLitMax >= MAX_LITERALS)
            EXIT_UNWANTED_STATE

        data->inst->isRelevantSolverLiteral[mapped] = 1;

        if (arguments.clauseFile)
        {
            fprintf(arguments.clauseFile, "%d ", mapped);
            printBasis(index2Base[i]);
            fprintf(arguments.clauseFile, "\n");
        }
    }

    (data->inst->solveLitMax)++; // Additional place because literals start at 1
    int max = data->inst->solveLitMax;
    if (max > MAX_LITERALS)
        EXIT_UNWANTED_STATE

    // memory and co for states. TODO in the future, lists could be independent from thread
    for (size_t i = 0; i < threads; i++)
    {
        data->states[i].assignment = malloc(sizeof(truth_value_t) * max);
        CHECK_OUT_OF_MEMORY(data->states[i].assignment)
        memset(data->states[i].assignment, 0, sizeof(truth_value_t) * max);
        data->states[i].sym_state.clauseList.clauses = malloc(sizeof(clause_t) * INITIAL_SIZE_CLAUSE_LIST);
        CHECK_OUT_OF_MEMORY(data->states[i].sym_state.clauseList.clauses);
        data->states[i].sym_state.clauseList.maxN = INITIAL_SIZE_CLAUSE_LIST;

        data->states[i].sym_state.ready2fire.elems = malloc(sizeof(int) * INITIAL_SIZE_INTEGER_LIST);
        CHECK_OUT_OF_MEMORY(data->states[i].sym_state.ready2fire.elems);
        data->states[i].sym_state.ready2fire.maxN = INITIAL_SIZE_INTEGER_LIST;

        // other lists are allocated when needed.
    }

    // Begin adding clauses
    if (arguments.clauseFile)
        fprintf(arguments.clauseFile, "clauses:\n");

    fflush(arguments.clauseFile);

    /*
        for A,B in permutations(N_choose_rank,2):
            for a in set(A)-set(B):
                CNF.append([-var_basis(A),-var_basis(B)]+[+var_basis(sorted_tuple(set(A)-{a}|{b})) for b in set(B)-set(A)])
    */
    // C version of this python code
    if (arguments.lazyEncoding == 0)
        for (int i = 0; i < N_BASES; i++)
            for (int j = 0; j < N_BASES; j++)
            {
                if (i == j)
                    continue;

                int clause[RANK * RANK + 2];
                clause[0] = -data->inst->baseIndex2LitSolver[i];
                clause[1] = -data->inst->baseIndex2LitSolver[j];
                int clauseSize = 2;

                int *base1 = index2Base[i];
                int *base2 = index2Base[j];

                int aSet[RANK], bSet[RANK];
                int aCount = 0;
                int bCount = 0;

                setminus(base1, base2, aSet, &aCount); // returns the indices and not the elements itself
                setminus(base2, base1, bSet, &bCount);

                if (aCount == 1) // just add element from other to get base 2
                    continue;

                for (int k = 0; k < aCount; k++)
                {
                    for (int l = 0; l < bCount; l++)
                    {
                        int swappedBase[RANK];
                        for (int m = 0; m < RANK; m++)
                            swappedBase[m] = base1[m];

                        swappedBase[aSet[k]] = base2[bSet[l]];

                        sortAfterChange(swappedBase, aSet[k], RANK);
                        int swappedBaseIndex = base2combinatorialIndex(swappedBase);
                        clause[clauseSize++] = data->inst->baseIndex2LitSolver[swappedBaseIndex];
                    }

                    bool r;
                    if (!clingo_propagate_init_add_clause(init, clause, clauseSize, &r))
                        EXIT_UNWANTED_STATE
                    if (arguments.clauseFile)
                    {
                        printClause(clause, clauseSize);
                        fprintf(arguments.clauseFile, ";b;%d\n", base1[aSet[k]]);
                    }

                    numberOfInitialClauses++;
                    clauseSize = 2;
                }
            }

    if (ADD_BINARY && arguments.lazyEncoding != 0)
        for (int i = 0; i < N_BASES; i++)
            for (int j = 0; j < N_BASES; j++)
            {
                if (i == j)
                    continue;

                int clause[RANK * RANK + 2];
                clause[0] = -data->inst->baseIndex2LitSolver[i];
                clause[1] = -data->inst->baseIndex2LitSolver[j];
                int clauseSize = 2;

                int *base1 = index2Base[i];
                int *base2 = index2Base[j];

                int aSet[RANK], bSet[RANK];
                int aCount = 0;
                int bCount = 0;

                setminus(base1, base2, aSet, &aCount); // returns the indices and not the elements itself
                setminus(base2, base1, bSet, &bCount);

                if (aCount != 2) // no binary
                    continue;

                printf("%d %d %d %d; %d %d %d %d \n", base1[0], base1[1], base1[2], base1[3], base2[0], base2[1], base2[2], base2[3]);

                for (int k = 0; k < aCount; k++)
                {
                    for (int l = 0; l < bCount; l++)
                    {
                        int swappedBase[RANK];
                        for (int m = 0; m < RANK; m++)
                            swappedBase[m] = base1[m];

                        swappedBase[aSet[k]] = base2[bSet[l]];

                        sortAfterChange(swappedBase, aSet[k], RANK);
                        int swappedBaseIndex = base2combinatorialIndex(swappedBase);
                        clause[clauseSize++] = data->inst->baseIndex2LitSolver[swappedBaseIndex];
                    }

                    bool r;
                    if (!clingo_propagate_init_add_clause(init, clause, clauseSize, &r))
                        EXIT_UNWANTED_STATE
                    if (arguments.clauseFile)
                    {
                        printClause(clause, clauseSize);
                        fprintf(arguments.clauseFile, ";b;%d\n", base1[aSet[k]]);
                    }

                    assert(clauseSize == 4);

                    numberOfInitialClauses++;
                    clauseSize = 2;
                }
            }

    // add with all the fixed bases.
    if (arguments.lazyEncoding != 0)
    {
        for (int i = 0; i < N_BASES; i++)
            for (int j = 0; j < N_BASES; j++)
            {

                if (i == j)
                    continue;

                if (!containsElement(fixedBaseIndices, RANK, i) && !containsElement(fixedBaseIndices, RANK, j))
                    continue;

                int clause[RANK * RANK + 2];
                clause[0] = -data->inst->baseIndex2LitSolver[i];
                clause[1] = -data->inst->baseIndex2LitSolver[j];
                int clauseSize = 2;

                int *base1 = index2Base[i];
                int *base2 = index2Base[j];

                int aSet[RANK], bSet[RANK];
                int aCount = 0;
                int bCount = 0;

                setminus(base1, base2, aSet, &aCount); // returns the indices and not the elements itself
                setminus(base2, base1, bSet, &bCount);

                if (aCount == 1) // just add element from other to get base 2
                    continue;

                for (int k = 0; k < aCount; k++)
                {
                    for (int l = 0; l < bCount; l++)
                    {
                        int swappedBase[RANK];
                        for (int m = 0; m < RANK; m++)
                            swappedBase[m] = base1[m];

                        swappedBase[aSet[k]] = base2[bSet[l]];

                        sortAfterChange(swappedBase, aSet[k], RANK);
                        int swappedBaseIndex = base2combinatorialIndex(swappedBase);
                        clause[clauseSize++] = data->inst->baseIndex2LitSolver[swappedBaseIndex];
                    }

                    bool r;
                    if (!clingo_propagate_init_add_clause(init, clause, clauseSize, &r))
                        EXIT_UNWANTED_STATE
                    if (arguments.clauseFile)
                    {
                        printClause(clause, clauseSize);
                        fprintf(arguments.clauseFile, ";b;%d\n", base1[aSet[k]]);
                    }
                    numberOfInitialClauses++;
                    clauseSize = 2;
                }
            }
    }

    int clause[RANK];
    bool r;

    // Last rows must be false
    if (arguments.maxTransversals == 0)
        arguments.maxTransversals = RANK - 1;

    for (int i = arguments.maxTransversals; i < RANK; i++)
    {
        int base[RANK];
        for (int j = 0; j < RANK; j++)
            base[j] = j * RANK + i;

        int index = base2combinatorialIndex(base);
        clause[0] = -data->inst->baseIndex2LitSolver[index];
        if (!clingo_propagate_init_add_clause(init, clause, 1, &r))
            EXIT_UNWANTED_STATE
        if (arguments.clauseFile)
        {
            printClause(clause, 1);
            fprintf(arguments.clauseFile, ";t\n");
        }
        numberOfInitialClauses++;
    }

    // first row always true
    {
        int base[RANK];

        for (int j = 0; j < RANK; j++)
            base[j] = j * RANK;
        int index = base2combinatorialIndex(base);
        clause[0] = data->inst->baseIndex2LitSolver[index];
        if (!clingo_propagate_init_add_clause(init, clause, 1, &r))
            EXIT_UNWANTED_STATE

        if (arguments.clauseFile)
        {
            printClause(clause, 1);
            fprintf(arguments.clauseFile, ";t\n");
        }
    }

    if (arguments.minTransversals == 0)
        arguments.minTransversals = 1;

    for (int i = 0; i < arguments.minTransversals; i++)
    {
        int base[RANK];

        for (int j = 0; j < RANK; j++)
            base[j] = j * RANK + i;
        int index = base2combinatorialIndex(base);
        clause[0] = data->inst->baseIndex2LitSolver[index];
        bool r;
        if (!clingo_propagate_init_add_clause(init, clause, 1, &r))
            EXIT_UNWANTED_STATE
        if (arguments.clauseFile)
        {
            printClause(clause, 1);
            fprintf(arguments.clauseFile, ";t\n");
        }
    }

    for (int i = 0; i < RANK; i++) // each column is a base
    {
        int base[RANK];
        for (int j = 0; j < RANK; j++)
        {
            base[j] = i * RANK + j;
        }
        int index = base2combinatorialIndex(base);
        clause[0] = data->inst->baseIndex2LitSolver[index];
        if (!clingo_propagate_init_add_clause(init, clause, 1, &r)) // add the unit clause
            EXIT_UNWANTED_STATE
        if (arguments.clauseFile)
        {
            printClause(clause, 1);
            fprintf(arguments.clauseFile, ";t\n");
        }
        numberOfInitialClauses++;
    }

    if (arguments.girth != 0)
    {
        int *clause = malloc(MAX_LITERALS);
        CHECK_OUT_OF_MEMORY(clause)

        int elements[RANK];
        pavingRec(elements, 0, init, data, clause, 0);
        free(clause);
    }

    data->inst->start = clock();

    printf("c\tnumber of initial clauses: %d\n", numberOfInitialClauses);
    fflush(stdout);
    return true;
}

void pavingRec2(int elements[RANK], int depth, int *clause, int *nLiterals, clingo_propagate_init_t *init, propagator_t *data, int min);

void pavingRec(int elements[RANK], int depth, clingo_propagate_init_t *init, propagator_t *data, int *clause, int min)
{
    if (depth == arguments.girth - 1)
    {
        int nLiterals = 0;
        pavingRec2(elements, depth, clause, &nLiterals, init, data, 0);
        bool r;
        if (!clingo_propagate_init_add_clause(init, clause, nLiterals, &r))
            EXIT_UNWANTED_STATE

        if (arguments.clauseFile)
        {
            printClause(clause, nLiterals);
            fprintf(arguments.clauseFile, ";g;[%d", elements[0]);
            for (int i = 1; i < arguments.girth - 1; i++)
                fprintf(arguments.clauseFile, ",%d", elements[i]);
            fprintf(arguments.clauseFile, "]\n");
        }
        return;
    }

    for (int i = min; i < RANK * RANK; i++)
    {
        elements[depth] = i;
        pavingRec(elements, depth + 1, init, data, clause, i + 1);
    }
}

void pavingRec2(int elements[RANK], int depth, int *clause, int *nLiterals, clingo_propagate_init_t *init, propagator_t *data, int min)
{
    if (depth == RANK)
    {
        int copyElements[RANK];
        for (int i = 0; i < RANK; i++)
            copyElements[i] = elements[i];
        bubbleSort(copyElements, RANK);

        int index = base2combinatorialIndex(copyElements);
        clause[(*nLiterals)++] = data->inst->baseIndex2LitSolver[index];
        return;
    }

    for (int i = min; i < RANK * RANK; i++)
    {
        if (containsElement(elements, depth, i))
            continue;

        elements[depth] = i;
        pavingRec2(elements, depth + 1, clause, nLiterals, init, data, i + 1);
    }
}

bool propagate_clingo(clingo_propagate_control_t *control, const clingo_literal_t *changes, size_t size, propagator_t *data)
{
    if (DEBUG2)
    {

        const clingo_assignment_t *a = clingo_propagate_control_assignment(control);
        int decisionLvl = clingo_assignment_decision_level(a);
        if (decisionLvl == 10)
        {
            printf("Number of positive literals: %ld\n", clingo_assignment_size(a));
        }
    }
    if (DEBUG)
    {
        printf("Literals added: ");
        for (size_t i = 0; i < size; i++)
            printf("%d ", changes[i]);
        printf("\n");
    }
    int thread = clingo_propagate_control_thread_id(control);
    data->states[thread].calls++;
    state_t *state = &data->states[thread];

    clock_t start_prop, end_prop;
    start_prop = clock();

    for (size_t i = 0; i < size; i++)
    {
        int lit = changes[i];
        if (DEBUG)
            printf("Lit: %d\n", lit);
        int positiveLit = ABS(lit);
        state->assignment[positiveLit] = lit > 0 ? truth_value_true : truth_value_false;

        if (arguments.lazyEncoding == 1)
        {

            if (lit > 0) // new base pairs to check base exchange property
                addNewClauses(&state->sym_state, lit, state->assignment, data);

            if (CLAUSE_VERSION2)
                addLiteral2(&state->sym_state, lit, state->assignment);
            else
                addLiteral(&state->sym_state, lit, state->assignment);
            if (CHECK)
            {
                if (CLAUSE_VERSION2)
                    checkState2(&state->sym_state, state->assignment, data);
                else
                    checkState(&state->sym_state, state->assignment, data);
            }
        }
    }

    bool r = true;
    if (arguments.lazyEncoding == 1)
    {
        if (CLAUSE_VERSION2)
        {
            int n = state->sym_state.ready2fire.n;
            int clauseIndices[n];
            for (int i = 0; i < n; i++)
                clauseIndices[i] = state->sym_state.ready2fire.elems[i];
            for (int j = 0; j < n && r; j++)
            {
                int index = clauseIndices[j];
                // printf("Ready to fire %d\n", index);

                clause_t *cl = &(state->sym_state.clauseList.clauses[index]);

                updateClause(&state->sym_state, cl, state->assignment, index);
                if (cl->indexFire == -1)
                    continue;

                int clause[cl->n + 2]; // the clause added to the solver
                clause[0] = -cl->b1;
                clause[1] = -cl->b2;

                for (int i = 0; i < cl->n; i++)
                    clause[i + 2] = cl->literals[i];

                // TODO check if really a unit clause
                if (DEBUG)
                {
                    printf("Added clause by base exchange proptery\n");
                }
                if (!clingo_propagate_control_add_clause(control, clause, (cl->n) + 2, clingo_clause_type_learnt, &r))
                    EXIT_UNWANTED_STATE
                if (arguments.clauseFile)
                {
                    printClause(clause, (cl->n) + 2);
                    fprintf(arguments.clauseFile, ";b;%d\n", cl->e);
                }
                state->addedClauses++;
            }
        }
        else
            for (int i = 0; (i < state->sym_state.ready2fire.n) && r; i++)
            {
                int index = state->sym_state.ready2fire.elems[i];
                // printf("Ready to fire %d\n", index);

                clause_t *cl = &(state->sym_state.clauseList.clauses[index]);

                int clause[cl->n + 2]; // the clause added to the solver
                clause[0] = -cl->b1;
                clause[1] = -cl->b2;

                for (int i = 0; i < cl->n; i++)
                    clause[i + 2] = cl->literals[i];

                // TODO check if really a unit clause
                if (DEBUG)
                {
                    printf("Added clause by base exchange proptery\n");
                }
                if (!clingo_propagate_control_add_clause(control, clause, (cl->n) + 2, clingo_clause_type_learnt, &r))
                    EXIT_UNWANTED_STATE
                if (arguments.clauseFile)
                {
                    printClause(clause, (cl->n) + 2);
                    fprintf(arguments.clauseFile, ";b;%d\n", cl->e);
                }
                state->addedClauses++;
            }
    }

    if (DEBUG2)
    {
        printf("Time adding and adapting clauses %f for %ld literals\n", ((double)clock() - start_prop) / CLOCKS_PER_SEC, size);
        fflush(stdout);
    }

    end_prop = clock();
    state->propTime += (end_prop - start_prop);

    if (r && state->calls > data->inst->arguments.frequency)
    {
        start_prop = clock();
        int permutations[RANK][RANK];
        for (int i = 0; i < RANK; i++)
            for (int j = 0; j < RANK; j++)
                permutations[i][j] = RANK * i + j;
        countCalls = 0;
        countEndNodes = 0;
        clock_t startMaxCheck = clock();
        r = checkMaximility(permutations, 0, 0, 0, state->assignment, control, data, 0);
        double timeMaxCheck = ((double)clock() - startMaxCheck) / CLOCKS_PER_SEC;

        if (timeMaxCheck > 1) // 1
        {
            state->calls = -20;
        }
        else if (timeMaxCheck > 0.2) // 0.2 seonds
        {
            state->calls = -5;
        }
        else if (timeMaxCheck > 0.1) // 0.1 seonds
        {
            state->calls = -2;
        }
        else
        {
            state->calls = -1;
        }

        if (r) // if maximality check was useless skip for a while
        {
            state->calls *= MAX(arguments.frequency, 1);
        }

        if (DEBUG2)
        {
            printf("Callse: %d; endnodes: %d, able to discard assignment %d\n", countCalls, countEndNodes, !r);
            printf("Number of clauses %d\n", state->sym_state.clauseList.n);
            fflush(stdout);
        }
        if (!r)
            state->addedClauses++;
        // printf("Maximality check added clause\n");

        if (DEBUG2)
        {
            printf("time minimality %f\n", timeMaxCheck);
            fflush(stdout);
        }

        end_prop = clock();
        state->maximilityCheckTime += (end_prop - start_prop);
    }

    // if (state->calls % 100000 == 0)
    // {
    //     //checkState(&state->sym_state, state->assignment, data); // TODO eventually delete but just to be sure.
    //     printf("c\tTotal vs prop time ratio: %f\n", ((double)state->time) / (clock() - data->inst->start));
    //     printf("c\tnumber of clauses %d\n", state->sym_state.clauseList.n);
    //     fflush(stdout);
    // }

    if (CHECK)
    {
        if (CLAUSE_VERSION2)
            checkState2(&state->sym_state, state->assignment, data);
        else
            checkState(&state->sym_state, state->assignment, data);
    }

    if (DEBUG2)
    {
        printf("Return from propagator\n");
        fflush(stdout);
    }

    return true;
}

bool check_basis_exchange(clingo_propagate_control_t *control, propagator_t *data)
{
    int thread = clingo_propagate_control_thread_id(control);
    truth_value_t *assignment = data->states[thread].assignment;

    int bases[N_BASES];
    int nBases = 0;

    for (int i = 0; i < N_BASES; i++)
    {
        if (assignment[data->inst->baseIndex2LitSolver[i]] == truth_value_true)
            bases[nBases++] = i;
    }

    for (int i1 = 0; i1 < nBases; i1++)
        for (int i2 = 0; i2 < nBases; i2++)
        {
            if (i1 == i2)
                continue;

            int *base1 = index2Base[bases[i1]];
            int *base2 = index2Base[bases[i2]];

            int aSet[RANK], bSet[RANK];
            int aCount = 0;
            int bCount = 0;

            setminus(base1, base2, aSet, &aCount); // returns the indices and not the elements itself
            setminus(base2, base1, bSet, &bCount);

            if (aCount == 1) // just add element from other to get base 2
                continue;

            for (int k = 0; k < aCount; k++)
            {
                int clause[3 * RANK];
                int clauseSize = 0;

                int fullfilled = false;
                for (int l = 0; l < bCount; l++)
                {
                    int swappedBase[RANK];
                    for (int m = 0; m < RANK; m++)
                        swappedBase[m] = base1[m];

                    swappedBase[aSet[k]] = base2[bSet[l]];

                    sortAfterChange(swappedBase, aSet[k], RANK);
                    int swappedBaseIndex = base2combinatorialIndex(swappedBase);
                    if (assignment[data->inst->baseIndex2LitSolver[swappedBaseIndex]] == truth_value_true)
                    {
                        fullfilled = true;
                        break;
                    }
                }

                if (!fullfilled)
                {
                    bool r;
                    if (!clingo_propagate_control_add_clause(control, clause, clauseSize, clingo_clause_type_static, &r))
                        EXIT_UNWANTED_STATE
                    if (r)
                        EXIT_UNWANTED_STATE
                    if (arguments.clauseFile)
                    {
                        printClause(clause, clauseSize);
                        fprintf(arguments.clauseFile, ";b;%d\n", base1[aSet[k]]);
                    }

                    numberOfInitialClauses++;
                    clauseSize = 2;
                    return false;
                }
            }
        }
    return true;
}

// Optional check at the end (i.e., if model was found.)
bool check_clingo(clingo_propagate_control_t *control, propagator_t *data)
{
    int thread = clingo_propagate_control_thread_id(control);

    clock_t start_prop, end_prop;
    start_prop = clock();

    int permutations[RANK][RANK];
    for (int i = 0; i < RANK; i++)
        for (int j = 0; j < RANK; j++)
        {
            permutations[i][j] = RANK * i + j;
        }

    end_prop = clock();
    data->states[thread].propTime += (end_prop - start_prop);

    start_prop = clock();
    int r = checkMaximility(permutations, 0, 0, 0, data->states[thread].assignment, control, data, 0);
    if (r && !check_basis_exchange(control, data)) // returns false, if a clause was added
    {
        data->states[thread].addedClauses++;
        if (data->states[thread].addedClauses == __LONG_MAX__)
            printf("Counter limit of added Clauses reached");
        r = false;
    }

    end_prop = clock();
    data->states[thread].maximilityCheckTime += end_prop - start_prop;
    return true;
}

bool propagate_undo(clingo_propagate_control_t *control, const clingo_literal_t *changes, size_t size, propagator_t *data)
{
    if (DEBUG)
    {
        printf("Literals undo: ");
        for (size_t i = 0; i < size; i++)
            printf("%d ", changes[i]);
        printf("\n");
    }
    int thread = clingo_propagate_control_thread_id(control);
    state_t *state = &data->states[thread];
    clock_t start_prop, end_prop;
    start_prop = clock();
    for (size_t i = 0; i < size; i++)
    {
        int lit = changes[i];
        state->assignment[ABS(lit)] = truth_value_free;
        if (DEBUG)
            printf("Lit (undo): %d\n", lit);

        if (arguments.lazyEncoding == 1)
        {
            if (lit > 0)
                removeClauses(&state->sym_state, lit);

            if (CLAUSE_VERSION2)
                removeLiteral2(&state->sym_state, lit, state->assignment);
            else
                removeLiteral(&state->sym_state, lit, state->assignment);
        }
    }
    if (CHECK)
    {
        if (CLAUSE_VERSION2)
            checkState2(&state->sym_state, state->assignment, data);
        else
            checkState(&state->sym_state, state->assignment, data);
    }

    end_prop = clock();
    state->propTime += end_prop - start_prop;
    return true;
}

/**
 * Get literals to the bases
 */
void init_base_lits(rotta_inst_t *inst, clingo_symbolic_atoms_t const *atoms)
{
    clingo_signature_t sig;
    clingo_symbolic_atom_iterator_t atoms_it, atoms_ie;

    if (!clingo_signature_create("b", RANK, true, &sig)) // signatur to get base literals
        EXIT_UNWANTED_STATE
    if (!clingo_symbolic_atoms_end(atoms, &atoms_ie))
        EXIT_UNWANTED_STATE

    if (!clingo_symbolic_atoms_begin(atoms, &sig, &atoms_it))
        EXIT_UNWANTED_STATE

    int elemens[RANK];
    bool equal;
    clingo_literal_t lit;
    clingo_symbol_t sym;
    // stop iteration if the end is reached
    if (!clingo_symbolic_atoms_iterator_is_equal_to(atoms, atoms_it, atoms_ie, &equal))
        EXIT_UNWANTED_STATE
    int baseCounter = 0;
    while (!equal)
    {
        baseCounter++;
        if (!clingo_symbolic_atoms_literal(atoms, atoms_it, &lit))
            EXIT_UNWANTED_STATE

        if (lit < 0)
        {
            printf("Error: Currently only support positive literals!!!!!!!!!");
            EXIT_UNWANTED_STATE
        }

        // get symbol
        if (!clingo_symbolic_atoms_symbol(atoms, atoms_it, &sym))
            EXIT_UNWANTED_STATE
        symbol2Base(sym, elemens);
        // printf("Base: ");
        // for (int i = 0; i < RANK; i++)
        //     printf("%d ", elemens[i]);
        // printf("\n");

        inst->baseIndex2Lit[base2combinatorialIndex(elemens)] = lit;

        // advance to the next placement atom
        if (!clingo_symbolic_atoms_next(atoms, atoms_it, &atoms_it))
            EXIT_UNWANTED_STATE
        if (!clingo_symbolic_atoms_iterator_is_equal_to(atoms, atoms_it, atoms_ie, &equal))
            EXIT_UNWANTED_STATE
    }

    if (baseCounter != N_BASES)
    {
        printf("Number of bases %d vs %d \n", baseCounter, N_BASES);
        EXIT_UNWANTED_STATE
    }
}

void setThreads(rotta_inst_t *inst, int threads)
{
    clingo_configuration_t *conf;
    clingo_id_t root_key, sub_key;
    // get the configuration object and its root key
    if (!clingo_control_configuration(inst->control, &conf))
        EXIT_UNWANTED_STATE
    if (!clingo_configuration_root(conf, &root_key))
        EXIT_UNWANTED_STATE

    if (!clingo_configuration_map_at(conf, root_key, "solver", &sub_key))
        EXIT_UNWANTED_STATE
    if (!clingo_configuration_array_at(conf, sub_key, 0, &sub_key))
        EXIT_UNWANTED_STATE
    if (!clingo_configuration_map_at(conf, sub_key, "threads", &sub_key))
        EXIT_UNWANTED_STATE
    char buff[10];
    sprintf(buff, "%d", threads);
    if (!clingo_configuration_value_set(conf, sub_key, buff))
        EXIT_UNWANTED_STATE
}

/**
 * Indices of result = set1 \ set2 with respect to set1.
 */
static inline void setminus(int set1[RANK], int set2[RANK], int result[RANK], int *nElems)
{
    *nElems = 0;
    for (int k = 0; k < RANK; k++)
    {
        if (!containsElement(set2, RANK, set1[k]))
            result[(*nElems)++] = k; // saves index not value itself
    }
}

// init according to inst arguments
void init(rotta_inst_t *inst)
{
    clingo_symbolic_atoms_t const *atoms;
    clingo_part_t parts[] = {{"base", NULL, 0}};
    size_t buffsize = 10000;
    char buff[buffsize];

    if (!clingo_control_new(arguments.optionsClingo, arguments.nOptions, NULL, NULL, 20, &(inst->control)))
        EXIT_UNWANTED_STATE

    sprintf(buff, "num(0..%d).", ELEMENTS - 1);
    if (!clingo_control_add(inst->control, "base", NULL, 0, buff))
        EXIT_UNWANTED_STATE

    if (RANK == 3)
    {
        if (!clingo_control_add(inst->control, "base", NULL, 0, "{b(I0,I1,I2)} :- num(I0), num(I1), num(I2), I0 < I1, I1 < I2."))
            EXIT_UNWANTED_STATE
        // if (!clingo_control_add(inst->control, "base", NULL, 0, "b(0,1,2). b(3,4,5). b(6,7,8). "))
        //    EXIT_UNWANTED_STATE
    }
    if (RANK == 4)
    {
        if (!clingo_control_add(inst->control, "base", NULL, 0, "{b(I0,I1,I2,I3)} :- num(I0), num(I1), num(I2),  num(I3), I0 < I1, I1 < I2, I2 < I3."))
            EXIT_UNWANTED_STATE
    }

    if (RANK == 5)
    {
        if (!clingo_control_add(inst->control, "base", NULL, 0, "{b(I0,I1,I2,I3,I4)} :- num(I0), num(I1), num(I2),  num(I3),  num(I4), I0 < I1, I1 < I2, I2 < I3, I3 < I4."))
            EXIT_UNWANTED_STATE
    }

    if (arguments.oriented)
    {

        if (RANK == 3)
        {
            if (!clingo_control_add(inst->control, "base", NULL, 0, "1{s(I0,I1,I2,-1); s(I0,I1,I2,0);  s(I0,I1,I2,1)}1 :-  num(I0), num(I1), num(I2), I0 != I1, I0 != I2, I1 != I2."))
                EXIT_UNWANTED_STATE

            if (!clingo_control_add(inst->control, "base", NULL, 0, "s(I1,I0,I2,-Q) :- s(I0,I1,I2,Q)."))
                EXIT_UNWANTED_STATE

            if (!clingo_control_add(inst->control, "base", NULL, 0, "s(I0,I2,I1,-Q) :- s(I0,I1,I2,Q)."))
                EXIT_UNWANTED_STATE

            if (!clingo_control_add(inst->control, "base", NULL, 0, ":- s(I0,I1,I2,0), b(I0,I1,I2), I0 < I1, I1 < I2."))
                EXIT_UNWANTED_STATE

            if (!clingo_control_add(inst->control, "base", NULL, 0, "b(I0,I1,I2) :- s(I0, I1, I2, Q), Q != 0, I0 < I1, I1 < I2."))
                EXIT_UNWANTED_STATE

            if (!clingo_control_add(inst->control, "base", NULL, 0, " :-  s(Y1, X2, X3, Q1), s(X1, Y2, X3, Q2), Q1 *Q2 >= 0, s(Y2, X2, X3, Q3), s(Y1, X1, X3, Q4), Q3 *Q4 >= 0, s(X1, X2, X3, Q5),  s(Y1, Y2, X3, Q6),  Q5 *Q6 < 0. "))
                EXIT_UNWANTED_STATE

            // if (!clingo_control_add(inst->control, "base", NULL, 0, "b(0,1,2). b(3,4,5). b(6,7,8). "))
            //    EXIT_UNWANTED_STATE
        }

        if (RANK == 4)
        {
            if (!clingo_control_add(inst->control, "base", NULL, 0, "1{s(I0,I1,I2,I3,-1); s(I0,I1,I2,I3,0);  s(I0,I1,I2,I3,1)}1 :-  num(I0), num(I1), num(I2), num(I3), I0 != I1, I0 != I2, I0 != I3, I1 != I2, I1 != I3, I2 != I3."))
                EXIT_UNWANTED_STATE

            if (!clingo_control_add(inst->control, "base", NULL, 0, "s(I1,I0,I2, I3,-Q) :- s(I0,I1,I2,I3,Q)."))
                EXIT_UNWANTED_STATE

            if (!clingo_control_add(inst->control, "base", NULL, 0, "s(I0,I2,I1,I3,-Q) :- s(I0,I1,I2,I3,Q)."))
                EXIT_UNWANTED_STATE

            if (!clingo_control_add(inst->control, "base", NULL, 0, "s(I0,I1,I3,I2,-Q) :- s(I0,I1,I2,I3,Q)."))
                EXIT_UNWANTED_STATE

            if (!clingo_control_add(inst->control, "base", NULL, 0, ":- s(I0,I1,I2,I3,0), b(I0,I1,I2,I3), I0 < I1, I1 < I2, I2 < I3."))
                EXIT_UNWANTED_STATE

            if (!clingo_control_add(inst->control, "base", NULL, 0, "b(I0,I1,I2,I3) :- s(I0,I1,I2,I3,Q), Q != 0, I0 < I1, I1 < I2, I2 < I3."))
                EXIT_UNWANTED_STATE

            if (!clingo_control_add(inst->control, "base", NULL, 0, " :-  s(Y1,X2,X3,X4,Q1), s(X1,Y2,X3,X4,Q2), Q1*Q2 >= 0, s(Y2,X2,X3,X4,Q3), s(Y1,X1,X3,X4,Q4), Q3*Q4 >= 0, s(X1,X2,X3,X4,Q5),  s(Y1,Y2,X3,X4,Q6),  Q5*Q6 < 0."))
                EXIT_UNWANTED_STATE
        }
    }

    printf("c\tStart grounding\n");
    fflush(stdout);

    if (!clingo_control_ground(inst->control, parts, 1, NULL, NULL))
        EXIT_UNWANTED_STATE

    printf("c\tFinished grounding\n");
    fflush(stdout);

    propagator_t *prop_data = malloc(sizeof(propagator_t));
    CHECK_OUT_OF_MEMORY(prop_data)
    memset(prop_data, 0, sizeof(propagator_t));
    prop_data->inst = inst;
    inst->propData = prop_data;

    clingo_propagator_t *prop = malloc(sizeof(clingo_propagator_t));
    CHECK_OUT_OF_MEMORY(prop)
    memset(prop, 0, sizeof(clingo_propagator_t));

    prop->init = (clingo_propagator_init_callback_t)init_clingo;
    prop->propagate = (clingo_propagator_propagate_callback_t)propagate_clingo;
    prop->undo = (clingo_propagator_undo_callback_t)propagate_undo;
    prop->decide = NULL;
    prop->check = (clingo_propagator_check_callback_t)check_clingo;

    // get the container for symbolic atoms
    if (!clingo_control_symbolic_atoms(inst->control, &atoms))
        EXIT_UNWANTED_STATE

    init_base_lits(inst, atoms);

    if (!clingo_control_register_propagator(inst->control, prop, prop_data, false))
        EXIT_UNWANTED_STATE
}

// TODO check if int is sufficient or we have to take long and test function!!!
static inline int binom(int n, int k)
{
    if (n == 0 && k == 0) // both zero
        return 1;
    if (n < k)
        return 0;
    int result = 1;
    for (int i = 0; i < k; i++)
        result *= n - i;
    for (int i = 2; i <= k; i++) // divison by k factorial
        result /= i;
    return result;
}

// See https://math.stackexchange.com/questions/349924/function-mapping-combinations-to-natural-numbers , i.e., combinatorial number system https://en.wikipedia.org/wiki/Combinatorial_number_system
static inline int base2combinatorialIndex(int base[RANK])
{
    int result = 0;
    for (int i = 0; i < RANK; i++)
        result += binom(base[i], i + 1);
    return result;
}

// --------------------------------clause update stuff---------------------------

// adds new clauses because with all new base pairs
void addNewClauses(clauses_state_t *state, lit_t lit, truth_value_t *assignment, propagator_t *data)
{
    if (DEBUG)
    {
        printf("ADD positive literal\n");
    }
    int i = data->inst->solverLit2baseIndex[lit];
    if (containsElement(fixedBaseIndices, RANK, i)) // Only relevant for the beginning because implied by BCP
        return;
    for (int j = 0; j < N_BASES; j++) // TODO maybe also save the present bases to avoid this. ERROR!!!!! if several are added because in both directions
    {
        if (j == i ||                                                           // same base
            containsElement(fixedBaseIndices, RANK, j) ||                       // already added in init_clingo
            assignment[data->inst->baseIndex2LitSolver[j]] != truth_value_true) // also second base must be present
            continue;

        int *bases[2][2] = {{index2Base[i], index2Base[j]}, {index2Base[j], index2Base[i]}}; // base exchange property in both directions for the two bases

        for (int c = 0; c < 2; c++)
        {
            int litB1 = lit;
            int litB2 = data->inst->baseIndex2LitSolver[j];

            int *base1 = bases[c][0];
            int *base2 = bases[c][1];

            int aSet[RANK], bSet[RANK];
            int aCount = 0;
            int bCount = 0;

            assert(aCount == bCount);

            setminus(base1, base2, aSet, &aCount); // returns the indices and not the elements itself
            setminus(base2, base1, bSet, &bCount);

            if (aCount == 1) // just add element from other to get base 2
                continue;

            if (ADD_BINARY && aCount == 2)
                continue;

            for (int k = 0; k < aCount; k++)
            {
                int literals[RANK];
                for (int l = 0; l < bCount; l++)
                {
                    int swappedBase[RANK];
                    for (int m = 0; m < RANK; m++)
                        swappedBase[m] = base1[m];

                    swappedBase[aSet[k]] = base2[bSet[l]];

                    sortAfterChange(swappedBase, aSet[k], RANK);
                    int swappedBaseIndex = base2combinatorialIndex(swappedBase);
                    literals[l] = data->inst->baseIndex2LitSolver[swappedBaseIndex];
                }
                if (CLAUSE_VERSION2)
                    addNewClause2(state, assignment, litB1, litB2, base1[aSet[k]], literals, bCount);
                else
                    addNewClause(state, assignment, litB1, litB2, base1[aSet[k]], literals, bCount);
            }
        }
    }
}

void removeClauses(clauses_state_t *state, lit_t lit)
{
    int_list_t *list = &state->baseLits2clauses[lit];
    while (list->n != 0)
    {
        // remove last element
        int index = list->elems[list->n - 1];
        if (CLAUSE_VERSION2)
            removeClause2(state, index);
        else
            removeClause(state, index);
    }
}

// -------------------------------------decision lvl cutoff staff------------------------------------------

// get the assigned literals
void getCurrentAssignedBases(int *clause, int *n, int *presentLiterals, int *nPresent, rotta_inst_t *inst, truth_value_t *assignment)
{
    for (int i = 0; i < inst->solveLitMax; i++)
    {
        if (!inst->isRelevantSolverLiteral[i])
            continue;

        switch (assignment[i])
        {
        case truth_value_false:
            clause[(*n)++] = i;
            presentLiterals[(*nPresent)++] = -i;
            break;

        case truth_value_true:
            clause[(*n)++] = -i;
            presentLiterals[(*nPresent)++] = i;
            break;

        default:
            break;
        }
    }
}